# Generated by Django 2.0 on 2018-01-26 08:51

from django.db import migrations, models
import django.db.models.deletion
import django.utils.timezone
import jutil.validators


class Migration(migrations.Migration):

    dependencies = [
        ('jacc', '0008_auto_20180121_1705'),
        ('jbank', '0010_auto_20171226_1013'),
    ]

    operations = [
        migrations.CreateModel(
            name='Refund',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', models.DateTimeField(blank=True, db_index=True, default=django.utils.timezone.now, editable=False, verbose_name='created')),
                ('iban_account_number', models.CharField(max_length=35, validators=[jutil.validators.iban_validator], verbose_name='IBAN')),
                ('refund_date', models.DateField(blank=True, db_index=True, default=None, null=True, verbose_name='refund date')),
                ('archive_identifier', models.CharField(blank=True, db_index=True, default='', max_length=32, verbose_name='archive identifier')),
                ('state', models.CharField(blank=True, db_index=True, default='W', max_length=1, verbose_name='state')),
                ('source', models.OneToOneField(on_delete=django.db.models.deletion.PROTECT, related_name='+', to='jacc.AccountEntry')),
            ],
            options={
                'verbose_name_plural': 'refunds',
                'verbose_name': 'refund',
            },
        ),
        migrations.AlterField(
            model_name='referencepaymentbatchfile',
            name='created',
            field=models.DateTimeField(blank=True, db_index=True, default=django.utils.timezone.now, editable=False, verbose_name='created'),
        ),
        migrations.AlterField(
            model_name='statementfile',
            name='created',
            field=models.DateTimeField(blank=True, db_index=True, default=django.utils.timezone.now, editable=False, verbose_name='created'),
        ),
    ]
