# -*- coding: utf-8 -*-
# Generated by Django 1.11 on 2017-11-02 23:41
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('jbank', '0005_statementfile_created'),
    ]

    operations = [
        migrations.AddField(
            model_name='statementfile',
            name='errors',
            field=models.TextField(blank=True, default='', max_length=4086),
        ),
        migrations.AlterField(
            model_name='statementfile',
            name='file',
            field=models.FileField(upload_to='uploads', verbose_name='file'),
        ),
    ]
