# Generated by Django 2.0.2 on 2018-04-11 18:14

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('jbank', '0021_auto_20180209_0935'),
    ]

    operations = [
        migrations.AlterField(
            model_name='referencepaymentbatch',
            name='service_identifier',
            field=models.CharField(blank=True, max_length=9, verbose_name='service identifier'),
        ),
    ]
